@@ -20,6 +20,9 @@ from bs4 import BeautifulSoup as _BeautifulSoup
 from math import ceil as _ceil
 import user_agent
 from .errors import *
+from multiprocessing import Process
+from multiprocessing.dummy import Pool
+from itertools import chain
 
 """ ERRORS EXCEPTIONS """
 
@@ -96,6 +99,10 @@ class Github:
 
         self.CURRENT_PATH = _os.getcwd()
 
+        # ---------------- CHECKING IF ACCOUNT IS ORGANIZATION OR USER ----------------
+        if str(self.__resp["type"]) == "Organization":
+            raise OrganizationAccount("Organization accounts is NOT associated with this package.")
+
         ##################################### Followers Response #####################################
 
         self.__followersResp = _BeautifulSoup(_get(f"https://github.com/{self.username}?tab=followers").text,
@@ -136,25 +143,27 @@ class Github:
         # except:
         #     self.blog = None
 
-    def getFollowersImgs(self):
+    def getFollowersImgs(self) -> list:
+        followersImgsLinksList = []
+
+        """ TESTING/DEBUGGING """
+        # followersUsernames = self.getFollowersUsernames()
+        # followersImgsLinksList = list(map(self.getProfileImg, followersUsernames))
+        """ TESTING/DEBUGGING """
         folloersImgsAttrs = {"class": "d-table table-fixed col-12 width-full py-4 border-bottom color-border-secondary"}
 
-        followersImgsLinksList = []
         for pageNum in range(1, _ceil(self.followers / 50) + 1):
-            if pageNum == 1:
-                soup = _BeautifulSoup(_get(f"https://github.com/{self.username}?&tab=followers").text,
-                                      "html.parser")
-            else:
-                soup = _BeautifulSoup(
-                    _get(f"https://github.com/{self.username}?page={str(pageNum)}&tab=followers").text,
-                    "html.parser")
+            # if pageNum == 1:
+            #     soup = _BeautifulSoup(_get(f"https://github.com/{self.username}?&tab=followers").text,
+            #                           "html.parser")
+            # else:
+            soup = _BeautifulSoup(
+                _get(f"https://github.com/{self.username}?page={str(pageNum)}&tab=followers").text,
+                "html.parser")
 
-            followersContainer = soup.findAll("div", attrs=folloersImgsAttrs)
-
-            for followerTag in followersContainer:
-                # print(followerTag.find("img", attrs={"class": "avatar avatar-user"})["src"][:-10])
-                followersImgsLinksList.append(
-                    followerTag.find("img", attrs={"class": "avatar avatar-user"})["src"][:-10])
+            followersImgsLinksList.extend(
+                [followerTag.find("img", attrs={"class": "avatar avatar-user"})["src"][:-10] for followerTag in
+                 soup.findAll("div", attrs=folloersImgsAttrs)])
 
         return followersImgsLinksList
 
@@ -226,7 +235,7 @@ class Github:
         ############################################## End User Info ##############################################
         return res
 
-    def countryinfo(self):
+    def countryInfo(self):
         """ A simple function to collect information about any country """
         if self.countryName is None:
             return None
@@ -237,18 +246,20 @@ class Github:
             return None
         return countryinfo
 
-    def downloadProfilePic(self, user=None, Path=current_path):
+    def downloadProfileImg(self, user=None, Path=current_path):
         """ Method for downloading specific  profile picture of Github user or Id
          You can specify the path for downloading the picture, default is current path.
          """
         if user is None:  # download object profile picture.
             user = self.username
-            picUrl = f"https://avatars.githubusercontent.com/u/{self.__resp['id']}?v=4"  # The profile pic of the
-            # object or instance.
-        else:  # download the profile picture of another user
-            picUrl = _BeautifulSoup(_get(f"https://github.com/{user}", headers=self.__headers).text,
-                                    "html.parser").find("img", attrs={
-                "class": "avatar avatar-user width-full border color-bg-primary"})["src"]
+        #     picUrl = f"https://avatars.githubusercontent.com/u/{self.__resp['id']}?v=4"  # The profile pic of the
+        #     # object or instance.
+        # else:  # download the profile picture of another user
+        #     picUrl = _BeautifulSoup(_get(f"https://github.com/{user}", headers=self.__headers).text,
+        #                             "html.parser").find("img", attrs={
+        #         "class": "avatar avatar-user width-full border color-bg-primary"})["src"]
+
+        picUrl = self.getProfileImg(user)
 
         self.__createPathAndEnterIn(Path)
         picName = user + ".jpeg"
@@ -270,6 +281,15 @@ class Github:
     #     pass
     """
 
+    def getProfileImg(self, user=None):
+        if user is None:
+            user = self.username
+        imgLink = _BeautifulSoup(
+            _get(f"https://github.com/{user}", headers={"User-Agent": user_agent.generate_user_agent()}).text,
+            "html.parser").find("img", attrs={"class": "avatar avatar-user width-full border color-bg-primary"})["src"][
+                  :-4]
+        return imgLink
+
     def getProfileById(self, UserId: str) -> str:
         """ Method that gets the user link profile by his Github id """
         api = f"https://api.github.com/user/{UserId}"  # API for return user info by ID.
@@ -321,34 +341,55 @@ class Github:
 
         return followingImgsLinksList
 
-    def downloadFollowersPics(self, Path=current_path):
-        """ Method for downloading followers profile phot_os  """
+    def __downloadUsrImgByUrl(self, username, imgUrl):
+        imgName = str(username) + ".jpeg"
+        _urlretrieve(imgUrl, imgName)
+
+    def downloadFollowersImgs(self, Path=current_path):
+        """ Method for downloading followers profile photos """
 
         self.__createPathAndEnterIn(Path)
 
         followersProfilesPicsList = self.getFollowersImgs()
         followersUsernames = self.getFollowersUsernames()
 
-        """ Downloading Profiles Pictures """
-        for imageLink, followerUsername in zip(followersProfilesPicsList, followersUsernames):
-            # imgName = (str(imageLink).split("/")[-1])[:-4] + ".jpeg"
-            imgName = followerUsername + ".jpeg"  # profile picture will be named with it's owner.
-            if not _pathlib.Path(imgName).exists():
-                _urlretrieve(imageLink, imgName)
+        """ Downloading Images Through multiprocessing/Threading """
+        processes = Pool(len(followersUsernames) // 5)
+        processes.starmap(self.__downloadUsrImgByUrl, zip(followersUsernames, followersProfilesPicsList))
+        processes.close()
+        processes.join()
+
+        """ TESTING/DEBUGGING """
+        """ Downloading Profiles Pictures/Images """
+        # for imageLink, followerUsername in zip(followersProfilesPicsList, followersUsernames):
+        #     # imgName = (str(imageLink).split("/")[-1])[:-4] + ".jpeg"
+        #     imgName = followerUsername + ".jpeg"  # profile picture will be named with it's owner.
+        #     if not _pathlib.Path(imgName).exists():
+        #         _urlretrieve(imageLink, imgName)
+
         _os.chdir(self.CURRENT_PATH)
 
-    def downloadFollowingPics(self, Path=current_path):
-        """ Method for downloading following profile phot_os  """
+    def downloadFollowingImgs(self, Path=current_path):
+        """ Method for downloading following profile photos """
+
         self.__createPathAndEnterIn(Path)
 
         followingsProfilesPicsList = self.getFollowingImgs()
         followingUsernames = self.getFollowingUsernames()
+
+        """ Downloading Images Through multiprocessing/Threading """
+        processes = Pool(len(followingUsernames) // 5)
+        processes.starmap(self.__downloadUsrImgByUrl, zip(followingUsernames, followingsProfilesPicsList))
+        processes.close()
+        processes.join()
+
+        """ TESTING/DEBUGGING """
         """ Downloading Profiles Pictures """
-        for imageLink, followingUsername in zip(followingsProfilesPicsList, followingUsernames):
-            # imgName = (str(imageLink).split("/")[-1])[:-4] + ".jpeg"
-            imgName = followingUsername + ".jpeg"  # profile picture will be named with it's owner.
-            if not _pathlib.Path(imgName).exists():
-                _urlretrieve(imageLink, imgName)
+        # for imageLink, followingUsername in zip(followingsProfilesPicsList, followingUsernames):
+        #     # imgName = (str(imageLink).split("/")[-1])[:-4] + ".jpeg"
+        #     imgName = followingUsername + ".jpeg"  # profile picture will be named with it's owner.
+        #     if not _pathlib.Path(imgName).exists():
+        #         _urlretrieve(imageLink, imgName)
 
         _os.chdir(self.CURRENT_PATH)
 
@@ -380,7 +421,7 @@ class Github:
     def userEmail(self, user=None):
         """
         :return email of a specific user if provided
-        otherwise returns the email of object
+        otherwise returns the email of object's user.
         """
         if user is None:
             user = self.username
@@ -400,8 +441,8 @@ class Github:
         if user is None:
             user = self.username
         emails = [emailTag["href"][7:] for emailTag in _BeautifulSoup(
-            _get(f"https://github.com/{user}", headers={"User-Agent": user_agent.generate_user_agent()}).text,
-            "html.parser").select("a[href ^= 'mailto:']")]
+            _get(f"https://github.com/{user}", headers=self.__headers).text,
+            "html.parser").select("a[href^='mailto:']")]
         return emails
 
     def extLinks(self, user=None) -> list:
@@ -421,37 +462,21 @@ class Github:
                                                                                                                        "class": "Link--primary"})]
         return extLinks
 
-    def getFollowersUsernames(self) -> list:
-        """ get all followers usernames """
-
-        followersUsernames = []
-
-        for pageNum in range(1, _ceil(self.followers / 50) + 1):
-            if pageNum == 1:
-                soup = _BeautifulSoup(
-                    _get(f"https://github.com/{self.username}?&tab=followers", headers=self.__headers).text,
-                    "html.parser")
-            else:
-                soup = _BeautifulSoup(
-                    _get(f"https://github.com/{self.username}?page={str(pageNum)}&tab=followers",
-                         headers=self.__headers).text,
-                    "html.parser")
-
-            for followerUsernameTag in soup.find_all("span", attrs={"class": "Link--secondary"}):
-                followersUsernames.append(followerUsernameTag.text)
-
-        return followersUsernames
-
     # def getFollowersUsernames(self):
     #     _Thread(target=self._getFollowersUsernames).start()
 
-    def getFollowersEmails(self, output=None) -> list:
+    def getFollowersEmails(self, output=None) -> dict:
         """get all followers emails"""
 
         followersUsernames = self.getFollowersUsernames()
-        followersEmails = []
-        for followerUsername in followersUsernames:
-            followersEmails.append({followerUsername: self.userEmail(followerUsername)})
+        """ TESTING/DEBUGGING """
+        # followersEmails = []
+        # for followerUsername in followersUsernames:
+        #     followersEmails.append({followerUsername: self.userEmail(followerUsername)})
+
+        res = Pool(20).map(self.userEmail, followersUsernames)
+        #                               keys   , values
+        followersEmails = dict(zip(followersUsernames, res))
 
         # Saving emails inside an output file
         if output is not None:
@@ -460,33 +485,91 @@ class Github:
                     print(followerEmail, file=outputFile)
         return followersEmails
 
+    def __getFollowingUsernameByPageNum(self, pageNum) -> list:
+        soup = _BeautifulSoup(
+            _get(f"https://github.com/{self.username}?page={str(pageNum)}&tab=following",
+                 headers=self.__headers).text,
+            "html.parser")
+
+        return [followingUsernameTag.text for followingUsernameTag in
+                soup.find_all("span", attrs={"class": "Link--secondary"})]
+
+    def __getFollowerUsernameByPageNum(self, pageNum) -> list:
+        soup = _BeautifulSoup(
+            _get(f"https://github.com/{self.username}?page={str(pageNum)}&tab=followers",
+                 headers=self.__headers).text,
+            "html.parser")
+
+        return [followerUsernameTag.text for followerUsernameTag in
+                soup.find_all("span", attrs={"class": "Link--secondary"})]
+
     def getFollowingUsernames(self) -> list:
         """get all following usernames"""
 
         followingUsernames = []
 
-        for pageNum in range(1, _ceil(self.following / 50) + 1):
-            if pageNum == 1:
-                soup = _BeautifulSoup(
-                    _get(f"https://github.com/{self.username}?&tab=following", headers=self.__headers).text,
-                    "html.parser")
-            else:
-                soup = _BeautifulSoup(
-                    _get(f"https://github.com/{self.username}?page={str(pageNum)}&tab=following",
-                         headers=self.__headers).text,
-                    "html.parser")
+        """ TESTING/DEBUGGING """
+        # for pageNum in range(1, _ceil(self.following / 50) + 1):
+        #     # if pageNum == 1:
+        #     #     soup = _BeautifulSoup(
+        #     #         _get(f"https://github.com/{self.username}?&tab=following", headers=self.__headers).text,
+        #     #         "html.parser")
+        #     # else:
+        #     soup = _BeautifulSoup(
+        #         _get(f"https://github.com/{self.username}?page={str(pageNum)}&tab=following",
+        #              headers=self.__headers).text,
+        #         "html.parser")
 
-            for followingUsernameTag in soup.find_all("span", attrs={"class": "Link--secondary"}):
-                followingUsernames.append(followingUsernameTag.text)
+        # followingUsernames.extend(followingUsernameTag.text for followingUsernameTag in
+        #                           soup.find_all("span", attrs={"class": "Link--secondary"}))
+
+        pagesNum = list(range(1, _ceil(self.following / 50) + 1))
+        # for pageNum in pagesNum:
+        #     process = Process(target=self.__getFollowingUsernameByPageNum, args=(pageNum,)).start()
+
+        # An efficient way for unpacking list of tuples to list
+        res = list(chain.from_iterable(Pool(len(pagesNum)).map(self.__getFollowingUsernameByPageNum, pagesNum)))
+        followingUsernames.extend(res)
 
         return followingUsernames
 
-    def getFollowingEmails(self, output=None) -> list:
+    def getFollowersUsernames(self) -> list:
+        """ get all followers usernames """
+
+        followersUsernames = []
+
+        """ TESTING/DEBUGGING """
+        # for pageNum in range(1, _ceil(self.followers / 50) + 1):
+        #     if pageNum == 1:
+        #         soup = _BeautifulSoup(
+        #             _get(f"https://github.com/{self.username}?&tab=followers", headers=self.__headers).text,
+        #             "html.parser")
+        #     else:
+        #         soup = _BeautifulSoup(
+        #             _get(f"https://github.com/{self.username}?page={str(pageNum)}&tab=followers",
+        #                  headers=self.__headers).text,
+        #             "html.parser")
+        #
+        #     for followerUsernameTag in soup.find_all("span", attrs={"class": "Link--secondary"}):
+        #         followersUsernames.append(followerUsernameTag.text)
+
+        pagesNum = list(range(1, _ceil(self.followers / 50) + 1))
+        # An efficient way for unpacking list of tuples to list
+        res = list(chain.from_iterable(Pool(len(pagesNum)).map(self.__getFollowerUsernameByPageNum, pagesNum)))
+        followersUsernames.extend(res)
+
+        return followersUsernames
+
+    def getFollowingEmails(self, output=None) -> dict:
         """get all following emails"""
         followingUsernames = self.getFollowingUsernames()
-        followingEmails = []
-        for followingUsername in followingUsernames:
-            followingEmails.append({followingUsername: self.userEmail(followingUsername)})
+        """ TESTING/DEBUGGING """
+        # for followingUsername in followingUsernames:
+        #     followingEmails.append({followingUsername: self.userEmail(followingUsername)})
+
+        res = Pool(20).map(self.userEmail, followingUsernames)
+        #                               keys   , values
+        followingEmails = dict(zip(followingUsernames, res))
 
         # Saving emails inside an output file
         if output is not None:
